/*******************************************************************************
 * Copyright (c) 2011, 2012 SINTEF, Martin F. Johansen.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     SINTEF, Martin F. Johansen - the implementation
 *******************************************************************************/

package no.sintef.ict.splcatool.lasplt;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeoutException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.io.FileUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import splar.core.fm.FeatureModelException;

import de.ovgu.featureide.fm.core.io.UnsupportedModelException;

import no.sintef.ict.splcatool.CNF;
import no.sintef.ict.splcatool.CoveringArray;
import no.sintef.ict.splcatool.CoveringArrayFile;
import no.sintef.ict.splcatool.GUIDSL;
import no.sintef.ict.splcatool.SXFM;
import no.sintef.ict.splcatool.XLSXLib;

class FeatureData{
	String component;
	String version;
	String repository;
}

class TestData{
	String component;
	String version;
	String repository;
	public String testClass;
	public String testapp;
}

class TestResults{

	public Integer tests;
	public Integer failures;
	public Integer errors;
	public String time;
	public Integer success;
	public String timestamp;
	public long successrate;
	public String name;
	
}

class ExecutionResult{
	int retvalue;
	public boolean timeout;
	public int time;
}

public class LTSPLT {
	static public void main(String[] args) throws IOException, SecurityException, IllegalArgumentException, ClassNotFoundException, NoSuchMethodException, IllegalAccessException, InvocationTargetException, UnsupportedModelException, FeatureModelException, TimeoutException, InterruptedException{
		new LTSPLT().mainObj(args);
	}
	
	static final String memoptions = " -Xms40m -Xmx2g -XX:MaxPermSize=256m ";
	
	public void mainObj(String[] args) throws IOException, ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException, UnsupportedModelException, FeatureModelException, TimeoutException, InterruptedException{
		// Input files
		String baseDir = args[0];
		Integer t = new Integer(args[1]);
		Integer timeout = new Integer(args[2]);
		String packages_xlsx = baseDir + "/models/packages.xlsx";
		String testpackages_xlsx = baseDir + "/models/test-packages.xlsx";
		String featureModel = baseDir + "/models/Eclipse.m";
		String baseProductZip = baseDir + "/packages/eclipse-platform-3.7-win32-x86_64.zip";
		
		new FileUtility().writeStringToFile(baseDir + "/results/testparameters.txt", t+";"+timeout+";");
		
		// Generated files
		String packages = baseDir+"/results/packages.csv";
		String testpackages = baseDir+"/results/test-packages.csv";
		String cafile = baseDir+"/results/Eclipse.m.ca"+t+".csv";
		String productDir = baseDir+"/products";
		
		// Start
		// ***********************
		System.out.print("Converting Package Mapping File to CSV ");{
			/*if(new File(packages).exists()){
				System.out.println("[already exists]");
			}else{*/
				String csvstring = XLSXLib.getCSV(new File(packages_xlsx));
				new FileUtility().writeStringToFile(packages, csvstring);
				System.out.println("[done]");
			//}
		}
		
		// ***********************
		System.out.print("Converting Feature-Test Mapping File to CSV ");{
			/*if(new File(testpackages).exists()){
				System.out.println("[already exists]");
			}else{*/
				String csvstring = XLSXLib.getCSV(new File(testpackages_xlsx));
				new FileUtility().writeStringToFile(testpackages, csvstring);
				System.out.println("[done]");
			//}
		}
		
		// ***********************
		CSV packagescsv;
		CSV testpackagescsv;
		System.out.print("Loading Mapping Files ");{
			// Load test script
			packagescsv = new CSV(packages);
			testpackagescsv = new CSV(testpackages);
		}
		System.out.println("[done]");
		
		// ***********************
		System.out.print("Generating "+t+"-wise Covering Array ");{
			if(new File(cafile).exists()){
				System.out.println("[already exists]");
			}else{
				List<List<Integer>> sols;
				CoveringArray ca;
				
				ConsoleUtils.silenceOut();
				
				// Load Feature Model
				GUIDSL fm = new GUIDSL(featureModel);
				SXFM sxfm = fm.getSXFM();
				CNF cnf = sxfm.getCNF();
				
				// Generate Covering Array
				ca = cnf.getCoveringArrayGenerator("J11", 2, 1);
				ca.generate();
				
				sols = ca.getSolutionsAsList();
				System.out.println("Products: " + sols.size());
				
				ca.writeToFile(cafile);
				
				ConsoleUtils.restoreOut();
				System.out.println("[done]");
			}
		}
		
		// ***********************
		CoveringArray ca;
		System.out.print("Loading Covering Array ");{
			ca = new CoveringArrayFile(cafile);
			System.out.println("[done] Products: " + ca.getRowCount());
		}
		
		// ***********************
		//System.out.println("Test each product:");
		nextProduct:
		for(int i = 0; i < ca.getRowCount(); i++){
			//if(i!=3 && i!=4) continue; // Run product
			//if(i!=13) continue; // Run product
			
			System.out.println("Test Product " + i + ":");
			
			// ***********************
			System.out.print("\tInstalling base product: ");
			String currentProductDir = productDir + "/product" + i;
			if(new File(currentProductDir).exists()){
				System.out.println("[already exists]");
			}else{
				new File(currentProductDir).mkdir();
				String command = "7za x -o\"" + currentProductDir + "\" " + baseProductZip;
				//System.out.println("\n" + command);
				
				executeCommand(Integer.MAX_VALUE, command, baseDir, new StringBuffer());
				System.out.println("[done]");
			}
			
			// ***********************
			System.out.println("\tInstalling features: ");

			List<String> configuration = getFeatures(ca, i);
			
			Map<String, String> fs = null;
			for(String f : configuration){
				System.out.print("\t\tInstalling feature " + f + " ");
				FeatureData fd = getFeatureData(packagescsv, f);
				
				if(fd.component == null){
					System.out.println("[nothing to install]");
					continue;
				}
				
				if(fs == null){
					fs = getInstalledRoots(currentProductDir);
					fs.putAll(getInstalledFeatures(currentProductDir));
				}
				//System.out.println(fd.component);
				//System.out.println(fs);
				if(fs.containsKey(fd.component + ".feature.group")){ //  && fs.get(fd.component + ".feature.group").equals(fd.version)
					System.out.println("[already installed]");
					continue;
				}
				
				// Install
				//System.out.println(fd.component);
				
				//-application", "", "-repository", repository, "-installIU", component + "/" + version
				String command = "java -jar "+memoptions+" plugins/org.eclipse.equinox.launcher*.jar "+
					"-application org.eclipse.equinox.p2.director " +
					"-installIU " + fd.component + ".feature.group/" + fd.version + " " +
					"-repository " + ("file:///"+baseDir+"/"+fd.repository.replace(",", ","+"file:///"+baseDir+"/")) + " " + 
					"";
				//System.out.println(command);
				StringBuffer sb = new StringBuffer();
				ExecutionResult ret = executeCommand(Integer.MAX_VALUE, command, currentProductDir + "/eclipse", sb);
				
				if(ret.retvalue == 0){
					System.out.println("[done]");
				}else{
					System.out.println("[failed]");
					System.out.println(sb);
					continue nextProduct;
				}
				
				// reset cache
				fs = null;
			}
			
			// ***********************
			System.out.println("\tInstalling tests: ");
			
			fs = null;
			for(String f : configuration){
				System.out.println("\t\tInstalling tests for " + f + " ");
				
				List<TestData> tds = getTestsDataForFeature(testpackagescsv, f);
				
				for(TestData td : tds){
					System.out.print("\t\t\tInstalling test " + td.component + " ");
					
					if(fs == null){
						fs = getInstalledRoots(currentProductDir);
						//fs.putAll(getInstalledRoots(currentProductDir));
					}
					//System.out.println(fs);
					if(fs.containsKey(td.component) && fs.get(td.component).equals(td.version)){
						System.out.println("[already installed]");
						continue;
					}
					
					String command = "java -jar "+memoptions+" plugins/org.eclipse.equinox.launcher*.jar "+
						"-application org.eclipse.equinox.p2.director " +
						"-installIU " + td.component + "/" + td.version + " " +
						"-repository " + ("file:///"+baseDir+"/"+td.repository.replace(",", ","+"file:///"+baseDir+"/")) + " " + 
						"";
					
					StringBuffer sb = new StringBuffer();
					ExecutionResult ret = executeCommand(Integer.MAX_VALUE, command, currentProductDir + "/eclipse", sb);
					
					if(ret.retvalue == 0){
						System.out.println("[done]");
					}else{
						System.out.println("[failed]");
						System.out.println(sb);
						continue nextProduct;
					}
					
					// reset cache
					fs = null;
				}
			}
			
			//if(true) continue; // TODO: Skips testing this round
			
			// ***********************
			System.out.println("\tRunning tests: ");
			//Report r = new Report(baseDir + "/results/testReport.csv");
			//r.writeLine("Product;Feature;TestClass;Tests;Success;Failure;Error;Time;");
			for(String f : configuration){
				System.out.println("\t\tRunning tests for " + f + " ");
				
				List<TestData> tds = getTestsDataForFeature(testpackagescsv, f);
				
				for(TestData td : tds){
					if(td.testClass.trim().length()==0) continue;
					
					System.out.print("\t\t\tRunning test " + td.testClass + " ");
					
/*					System.out.println("\n");
					if(!td.testClass.equals("org.eclipse.ua.tests.AllTests")) continue; // Run this test suite
*/					
					String testResultFile = baseDir+"/results/product"+i+"-"+td.testClass+"-test-result.xml";
					
					boolean hadtimeout = false;
					boolean failed = false;
					boolean emptyResult = false;
					
					if(!new File(testResultFile).exists() || new File(testResultFile).length()==0 || true){ // don't skip
						
						// Delete workspace
						FileUtils.deleteDirectory(new File(currentProductDir + "/workspace."+td.testClass));
						
						String command = "java -jar "+memoptions+" plugins/org.eclipse.equinox.launcher*.jar "+
							"-application " + td.testapp + " " +
							"-os win32 " +
							"-ws win32 " +
							"-arch x86_64 " +
							"-dev bin " +
							"-testpluginname " + td.component + " " +
							"-classname " + td.testClass + " " +
							"-data " + currentProductDir + "/workspace."+td.testClass+" " +
							"formatter=org.apache.tools.ant.taskdefs.optional.junit.XMLJUnitResultFormatter,"+testResultFile+" " +
							"-consolelog" +
							"";
						
						//System.out.println(command);
						
						StringBuffer sb = new StringBuffer();
						ExecutionResult ret = executeCommand(timeout, command, currentProductDir + "/eclipse", sb);
						
						if(ret.timeout == true){
							System.out.println("[timeout]");
							hadtimeout = true;
						}else if(ret.retvalue != 0){
							System.out.println("[failed]");
							failed = true;
						}else if(new File(testResultFile).length()==0){
							System.out.println("[empty result]");
							emptyResult = true;
						}else if(ret.retvalue == 0){
							System.out.print("[done]");
						}
						
						// Write execution log
						new FileUtility().writeStringToFile(baseDir + "/results/failure-"+i+"-"+td.testClass+".log", sb.toString());
					}else{
						System.out.print("[already exists]");
					}
					
					// Read results
					if(!hadtimeout && !failed && !emptyResult){
						TestResults tr = getTestResults(testResultFile);
						if(tr != null){
							//r.writeLine(i + ";" + f + ";" + td.testClass + ";" + tr.tests + ";" + tr.success + ";" + tr.failures + ";" + tr.errors + ";" + tr.time + ";");
							System.out.println(" " + tr.successrate + "% ("+tr.success+"/"+tr.tests+")");
						}
					}
					
					//break;
				}
			}
			
			//break;
		}
	}
	static TestResults getTestResults(String testResultFile) {
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = null;
		try {
			db = dbf.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			//e.printStackTrace();
			return null;
		}
		Document dom = null;
		try {
			dom = db.parse(testResultFile);
		} catch (SAXException e) {
			e.printStackTrace();
			//System.out.println("Error parsing (SAXException) " + testResultFile);
			//System.exit(-1);
			return null;
		} catch (IOException e) {
			//e.printStackTrace();
			//System.exit(-1);
			return null;
		}
		
		Element docEle = dom.getDocumentElement();
		//NodeList nl = docEle.getElementsByTagName("testsuite");
		//Node item = nl.item(i).getAttributes().getNamedItem("tests");
		TestResults tr = new TestResults();
		tr.tests = new Integer(docEle.getAttribute("tests"));
		tr.failures = new Integer(docEle.getAttribute("failures"));
		tr.errors = new Integer(docEle.getAttribute("errors"));
		tr.time = docEle.getAttribute("time");
		tr.success = tr.tests - tr.failures - tr.errors;
		tr.timestamp = docEle.getAttribute("timestamp");
		tr.successrate = Math.round(((double)tr.success/tr.tests*100));
		tr.name = docEle.getAttribute("name");
		return tr;
	}
	static List<TestData> getTestsDataForFeature(CSV testpackagescsv, String f) {
		List<TestData> lfd = new ArrayList<TestData>();
		
		for(int i = 0; i < testpackagescsv.getHeight(); i++){
			if(!testpackagescsv.getColumnEntry("Feature", i).equals(f)) continue;
			//System.out.print(name);
			
			TestData td = new TestData();
			td.component = testpackagescsv.getColumnEntry("Test Component", i);
			td.version = testpackagescsv.getColumnEntry("Test Version", i);
			td.repository = testpackagescsv.getColumnEntry("Test Repository", i);
			td.testClass = testpackagescsv.getColumnEntry("Test Class", i);
			td.testapp = testpackagescsv.getColumnEntry("Test Application", i);
			lfd.add(td);
		}
		
		return lfd;
	}
	private FeatureData getFeatureData(CSV packagescsv, String featureName) {
		FeatureData fd = new FeatureData();
		
		for(int x = 0; x < packagescsv.getHeight(); x++){
			if(packagescsv.getColumnEntry("Feature", x).equals(featureName)){
				int rownr = x;
				if(!packagescsv.getColumnEntry("Component", rownr).equals("")){
					fd.component = packagescsv.getColumnEntry("Component", rownr);
					fd.version = packagescsv.getColumnEntry("Version", rownr);
					fd.repository = packagescsv.getColumnEntry("Repository", rownr);
				}
			}
		}
		
		return fd;
	}
	static List<String> getFeatures(CoveringArray ca, int i) {
		List<List<Integer>> sols = ca.getSolutionsAsList();
		List<String> configuration = new ArrayList<String>();
		for(int j = 0; j < sols.get(i).size(); j++){
			if(sols.get(i).get(j) > 0)
				configuration.add(ca.getId(sols.get(i).get(j)));
		}
		return configuration;
	}
	
	private Map<String, String> getInstalledFeatures(String currentProductDir) throws IOException, InterruptedException {
		String command = "java -jar "+memoptions+" plugins/org.eclipse.equinox.launcher*.jar -application org.eclipse.update.core.standaloneUpdate -command listFeatures";
		
		StringBuffer sb = new StringBuffer();
		executeCommand(Integer.MAX_VALUE, command, currentProductDir + "/eclipse", sb);
		String res = sb.toString();
		
		Map<String, String> map = new HashMap<String, String>();
		
		for(String line : res.split("\n")){
			String es[] = line.trim().split("[ ]+");
			if(es.length != 4) continue;
			if(es[0].equals("Feature:")){
				map.put(es[1], es[2]);
			}
		}
		
		return map;
	}

	private Map<String, String> getInstalledRoots(String currentProductDir) throws IOException, InterruptedException {
		String command = "java -jar "+memoptions+" plugins/org.eclipse.equinox.launcher*.jar -application org.eclipse.equinox.p2.director -listInstalledRoots";
		
		StringBuffer sb = new StringBuffer();
		executeCommand(Integer.MAX_VALUE, command, currentProductDir + "/eclipse", sb);
		String res = sb.toString();
		
		Map<String, String> map = new HashMap<String, String>();
		
		for(String line : res.split("\n")){
			String es[] = line.trim().split("/");
			if(es.length != 2) continue;
			map.put(es[0], es[1]);
		}
		
		return map;
	}
	
	static ExecutionResult executeCommand(long timeout, String command, String dir, StringBuffer sb) throws IOException, InterruptedException{
		ExecuteCommand ec = new ExecuteCommand(timeout, command, dir, sb);
		
		Thread t = new Thread(ec);
		t.start();
		t.join(timeout*1000 + 15*1000); // Wait 15 seconds extra
		
		t.interrupt();
		
		return ec.getResult();
	}
	
	/*static ExecutionResult executeCommand(int timeout, String command, String dir, StringBuffer sb) throws IOException, InterruptedException{
		Runtime rt = Runtime.getRuntime();
		Process p = rt.exec(command, null, new File(dir));
		InputStream inp = p.getInputStream();
		InputStream err = p.getErrorStream();
		
		ExecutionResult ret = new ExecutionResult();
		ret.retvalue = 0;
		ret.timeout = false;
		ret.time = 0;
		
		while(true){
			//System.out.println("waiting ... " + inp.available() + ", " + err.available());
			
			if((long)ret.time > (long)timeout*1000){
				ret.retvalue = 0;
				ret.timeout = true;
				p.destroy();
				Thread.sleep(2000);
			}
			
			while(inp.available() > 0){
				//System.out.print((char)inp.read());
				sb.append((char) inp.read());
			}
			
			while(err.available() > 0){
				//System.err.print((char)err.read());
				sb.append((char) err.read());
				
			}
			
			try{
				ret.retvalue = p.exitValue();
			}catch(IllegalThreadStateException e){
				Thread.sleep(100);
				ret.time += 100;
				continue;
			}

			break;
		}
		
		return ret;
	}*/
}
