package cvl.runtime;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.xmi.XMIResource;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;
import org.osgi.framework.Bundle;

import cvl.CVLModel;

public class CVLExecutionManager {

	public Bundle bundle = Platform.getBundle("org.variabilitymodeling.cvl1.2.transformations");
	final String METAMODEL_PATH = "/metamodels/UMLAndProfile.ecore";
	public URL u = bundle.getResource("/transformations/CVL1.2Transformation.m2t");
	protected String transformationPath = "";
	public GenerateRepresentation representation;
	boolean successful;
	Map<EObject, EObject> umlCorrTable;
	boolean useUML=false;
	String cvlPath="";
	String umlPath="";
	int numberOfResolutions=0;
	String inputCVLModel="";

	
	public List<String> getErrorList(){
		return representation.errorMsg;
	}
	
	public String getBaseExt(){
		return representation.baseExt;
	}
	
	public boolean isUseUML() {
		return useUML;
	}
	
	public boolean isSuccessful() {
		return successful;
	}
	
	public List<EObject> executeCVLModel(CVLModel cm){
		
		File tempInputFile;
		try {
			tempInputFile = File.createTempFile("CVLInput", ".cvl");
		} catch (IOException e) {
			MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to create temporary files.");
			successful = false;
			return null;
		}
		
		tempInputFile.deleteOnExit();

		ResourceSet resourceSet = new ResourceSetImpl();
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("*", new XMIResourceFactoryImpl());

		Resource resource = resourceSet.createResource(URI.createFileURI(tempInputFile.getAbsolutePath()));
		resource.getContents().add(cm);

		try {
			resource.save(Collections.EMPTY_MAP);
		} catch (IOException e) {
			MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to save temporary files.");
			return null;
		}

		initializeCVLModelFromFile(tempInputFile.getAbsolutePath());
		
		//store base models (one for each resolution)
		List<EObject> baseModels = new ArrayList<EObject>();
		
		int max = cm.getResolutionSpecification().size();
		
		File tempOutputFile;
		try {
			String ext = "xmi";
			if (isUseUML()){
				ext = "uml";
			}else{
				if (representation.baseExt.length() > 0){
					ext = representation.baseExt;
				}
			}
			tempOutputFile = File.createTempFile("CVLOutput", "." + ext);
			
		} catch (IOException e) {
			MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to create temporary files.");
			successful = false;
			return null;
		}
		tempOutputFile.deleteOnExit();
		
		executeCVLModelFromFile(tempOutputFile.getAbsolutePath());
		
		String fn = tempOutputFile.getAbsolutePath();
		String fnExt = fn.substring(fn.lastIndexOf("."));
		fn = fn.substring(0, fn.lastIndexOf("."));
		
		for (int i=0; i < max; i++){
			String tempName = fn + i + fnExt;

			ResourceSet newResourceSet = new ResourceSetImpl();
			newResourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(Resource.Factory.Registry.DEFAULT_EXTENSION, new XMIResourceFactoryImpl());

			//Load the CVL Model
			URI tempURI = URI.createFileURI(tempName);
			newResourceSet.createResource(tempURI);
			Resource newResource = newResourceSet.getResource(tempURI, false);
			try {
				newResource.load(Collections.EMPTY_MAP);
			} catch (IOException e) {
				MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to load temporary files.");
				return null;
			}

			//get root object
			Object root = newResource.getContents().get(0);

			if (root instanceof EObject){
				EObject eObj = (EObject) root;
				baseModels.add(eObj);
			}
		}

		return baseModels;
	}
	
	public void initializeCVLModelFromFile(String fileName){
		successful = false;
		useUML = false;
		try {
			u = FileLocator.resolve(u);
		} catch (IOException e) {
			return;
		}

		IPath p = new Path(new File(u.getFile()).getAbsolutePath()); 
		transformationPath = p.toOSString();

		//check if this is a UML model with profiles applied
		checkAndHandleUML(fileName);

		if (useUML){
			fileName = cvlPath;
		}

		representation = new GenerateRepresentation(fileName, transformationPath);
		inputCVLModel = fileName;
	}
	
	public void executeCVLModelFromFile(String path){
		
		if (path != null){
			String tempPath = path;
			if (useUML){
				File umlFile;
				try {
					umlFile = File.createTempFile("CVL", ".UMLAndProfile");
				} catch (IOException e) {
					MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to create temporary UML file.");
					successful = false;
					return;
				}
				tempPath = umlFile.getAbsolutePath();
				umlFile.delete();

			}

			System.setProperty("cvl_filename", tempPath);	

			//execute transformation
			String filePath = inputCVLModel.substring(0, inputCVLModel.lastIndexOf("\\"));
			representation.execute(filePath); //path as arg
			successful=true;


			if (useUML){
				reverseUML(tempPath, path);
			}
		}
	}
	
	private void reverseUML(String tempFile, String finalPath) {
		String suffix = ".UMLAndProfile";
		int index = tempFile.lastIndexOf(suffix);
		String fileNameBase = tempFile.substring(0, index);

		String finalSuffix = ".uml";
		int finalIndex = finalPath.lastIndexOf(finalSuffix);
		String finalFileNameBase = finalPath.substring(0, finalIndex);

		int i=0;
		String fileName = fileNameBase + i + suffix;
		String finalFileName = finalFileNameBase + i + finalSuffix;

		File file = new File(fileName);
		for (; i < numberOfResolutions;) {
			if (file.exists()){

				//workaround to fix the schema-location tag (is not properly generated by MOFScript)
				File origFile = new File(umlPath);
				String output="";
				try {
					Scanner origScanner = new Scanner(origFile);
					Scanner scanner = new Scanner(file);

					output += origScanner.nextLine();
					output += origScanner.nextLine();
					scanner.nextLine();
					scanner.nextLine();
					while (scanner.hasNextLine()) { 
						output += scanner.nextLine(); 
					} 
				} catch (FileNotFoundException e1) {
					e1.printStackTrace();
				}

				try {
					FileWriter outFile = new FileWriter(file);
					PrintWriter out = new PrintWriter(outFile);
					out.print(output);
					out.close();
					outFile.close();
				} catch (IOException e1) {
					e1.printStackTrace();
				}


				URI fileNameURI = URI.createFileURI(fileName);
				ResourceSet tempResourceSet = new ResourceSetImpl();
				tempResourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(Resource.Factory.Registry.DEFAULT_EXTENSION, new XMIResourceFactoryImpl());

				tempResourceSet.createResource(fileNameURI);
				Resource tempResource = tempResourceSet.getResource(fileNameURI, false);


				try {
					tempResource.load(Collections.EMPTY_MAP);
				} catch (IOException e) {
					MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to load temporary UML file.");
					successful = false;
					return;
				}

				List<?> clones = (List<?>) EcoreUtil.copyAll(tempResource.getContents());

				EObject newRoot = (EObject) clones.get(0);

				ResourceSet finalResourceSet = new ResourceSetImpl();
				finalResourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("*", new XMIResourceFactoryImpl());

				Resource finalResource = finalResourceSet.createResource(URI.createFileURI(finalFileName));

				for (EReference eRef : newRoot.eClass().getEAllContainments()){
					if (eRef.getUpperBound() == 1){ //uml model
						EObject eObj = (EObject) newRoot.eGet(eRef);
						finalResource.getContents().add(eObj);
					}else{ //profiles
						List profileList = (List) newRoot.eGet(eRef);
						for (Object object : profileList) {
							EObject eObj = (EObject) object;
							finalResource.getContents().add(eObj);	
						}
					}
				}

				Map options = new HashMap();
				options.put(XMIResource.OPTION_SCHEMA_LOCATION, Boolean.TRUE);

				try {
					finalResource.save(options);
				} catch (IOException e) {
					MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to save temporary CVL file.");
					return;
				}

				i++;
				fileName = fileNameBase + i + suffix;
				finalFileName = finalFileNameBase + i + finalSuffix;
				file = new File(fileName);
			}
		}
	}

	private void checkAndHandleUML(String fileName) {
		URI fileNameURI = URI.createFileURI(fileName);
		ResourceSet resourceSet = new ResourceSetImpl();
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(Resource.Factory.Registry.DEFAULT_EXTENSION, new XMIResourceFactoryImpl());

		resourceSet.createResource(fileNameURI);
		Resource resource = resourceSet.getResource(fileNameURI, false);
		try {
			resource.load(Collections.EMPTY_MAP);
		} catch (IOException e) {
			MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to load CVL file.");
			return;
		}

		Object mObj = resource.getContents().get(0);

		if (mObj instanceof CVLModel){ 
			CVLModel cm = (CVLModel) mObj;

			EObject base = cm.getBase();

			//use string instead of imported class to avoid having dependency to UML plugins
			if (base != null && base.getClass().getName().contains("uml") && base.eClass().getName().equals("Model")){ 
				if (base.eResource().getContents().size() > 1){
					System.out.println("UML with profiles detected...");
					useUML = true;
					createCombinedUMLModel(resource, base.eResource());
				}
			}
		}
	}




	private void createCombinedUMLModel(Resource cvlResource, Resource baseResource){
		Resource mmResource = loadMetamodelResource();

		if (mmResource == null){
			return;
		}

		Object obj = mmResource.getContents().get(0);

		if (obj instanceof EPackage){
			EPackage ePackage = (EPackage) obj;

			EObject rootObj = populateCombinedModel(ePackage, baseResource);

			//put the package UMLAndProfile into the package registry
			EPackage.Registry reg = EPackage.Registry.INSTANCE;
			reg.put(ePackage.getNsURI(), ePackage);

			CVLModel cmCopy = cloneCVLModel(cvlResource, rootObj);

			//save temporary files with the CVL and CombinedUML models
			File cvlFile=null;
			File baseFile=null;
			try {
				cvlFile = File.createTempFile("CVL", ".cvl");
				baseFile = File.createTempFile("CombinedUML", ".UMLAndProfile");
			} catch (IOException e1) {
				MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to create temp files.");
				return;
			}
			cvlFile.deleteOnExit();
			baseFile.deleteOnExit();

			ResourceSet newBaseResourceSet = new ResourceSetImpl();
			newBaseResourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("*", new XMIResourceFactoryImpl());

			Resource newBaseResource = newBaseResourceSet.createResource(URI.createFileURI(baseFile.getAbsolutePath()));
			newBaseResource.getContents().add(rootObj);

			Map options = new HashMap();
			options.put(XMIResource.OPTION_SCHEMA_LOCATION, Boolean.TRUE);

			try {
				newBaseResource.save(options);
			} catch (IOException e) {
				MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to save temporary CVL file.");
				return;
			}

			ResourceSet newCVLResourceSet = new ResourceSetImpl();
			newCVLResourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("*", new XMIResourceFactoryImpl());

			Resource newCVLResource = newCVLResourceSet.createResource(URI.createFileURI(cvlFile.getAbsolutePath()));
			newCVLResource.getContents().add(cmCopy);

			try {
				newCVLResource.save(Collections.EMPTY_MAP);
			} catch (IOException e) {
				MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to save temporary base model file.");
				return;
			}

			cvlPath = cvlFile.getAbsolutePath();
			umlPath = baseFile.getAbsolutePath();
		}

	}

	private CVLModel cloneCVLModel(Resource cvlResource, EObject rootObj) {

		List<?> clones = (List<?>) EcoreUtil.copyAll(cvlResource.getContents());
		CVLModel cmCopy = (CVLModel) clones.get(0);

		numberOfResolutions = cmCopy.getResolutionSpecification().size();

		Iterator<EObject> copyIt = EcoreUtil.getAllContents(clones);

		while (copyIt.hasNext()){
			EObject e = copyIt.next();

			for (EReference eRef : e.eClass().getEAllReferences()){

				//redirect pointer to Container of both model and profiles
				if (e instanceof CVLModel && eRef.getName().equals("base")){
					e.eSet(eRef, rootObj);
				}else{

					if (eRef.getUpperBound() == 1){
						Object obj = e.eGet(eRef);
						//We only want to redirect external references.
						//The copied CVL elements are not yet in a resource.
						if (obj instanceof EObject && ((EObject) obj).eResource() != null){
							if (umlCorrTable.containsKey(obj)){
								e.eSet(eRef, umlCorrTable.get(obj));
							}
						}
					}else{
						EList list = (EList)e.eGet(eRef);
						Map<EObject, Integer> toReplace = new HashMap<EObject, Integer>();

						for (Object obj : list) {
							if (obj instanceof EObject && ((EObject) obj).eResource() != null){
								if (umlCorrTable.containsKey(obj)){
									toReplace.put((EObject) obj, list.indexOf(obj));
								}
							}
						}

						for (EObject eObj : toReplace.keySet()) {
							list.remove(eObj);
							list.add(toReplace.get(eObj), umlCorrTable.get(eObj));
						}
					}
				}
			}
		}

		return cmCopy;
	}

	private Resource loadMetamodelResource() {
		URL mmURL = bundle.getResource(METAMODEL_PATH);

		try {
			mmURL = FileLocator.resolve(mmURL);
		} catch (IOException e) {
			return null;
		}

		IPath p = new Path(new File(mmURL.getFile()).getAbsolutePath()); 
		String path = p.toOSString();
		URI pathURI = URI.createFileURI(path);

		ResourceSet mmResourceSet = new ResourceSetImpl();
		mmResourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(Resource.Factory.Registry.DEFAULT_EXTENSION, new XMIResourceFactoryImpl());

		mmResourceSet.createResource(pathURI);
		Resource mmResource = mmResourceSet.getResource(pathURI, false);
		try {
			mmResource.load(Collections.EMPTY_MAP);
		} catch (IOException e) {
			MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "CVL", "Was not able to load metamodel for combined UML. Please check that the version of CVL is up to date.");
			return null;
		}

		return mmResource;
	}

	private Map<EObject,EObject> createCorrespondance(List<EObject> orig, List<?> copy){
		Map<EObject, EObject> corr = new HashMap<EObject, EObject>();

		Iterator<EObject> origIt = EcoreUtil.getAllContents(orig);
		Iterator<EObject> copyIt = EcoreUtil.getAllContents(copy);

		while (origIt.hasNext()){
			corr.put(origIt.next(), copyIt.next());
		}

		return corr;
	}

	private EObject populateCombinedModel(EPackage ePackage, Resource baseResource){
		EFactory instanceFactory = ePackage.getEFactoryInstance();

		EClass root = (EClass) ePackage.getEClassifiers().get(0);
		EObject rootObj = instanceFactory.create(root);

		EReference model = root.getEAllReferences().get(0);
		EReference profile = root.getEAllReferences().get(1);

		List<?> clones = (List<?>) EcoreUtil.copyAll(baseResource.getContents());

		umlCorrTable = createCorrespondance(baseResource.getContents(), clones);

		rootObj.eSet(model, clones.get(0));

		List<?> profileList = clones;
		profileList.remove(0);

		for (Object object : profileList){
			((List) rootObj.eGet(profile)).add(object);
		}

		return rootObj;
	}

}
