package cvl.runtime;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.domain.EditingDomain;

import cvl.CVLModel;
import cvl.CompositeVariability;
import cvl.CvlFactory;
import cvl.CvlPackage;
import cvl.ExecutablePrimitive;
import cvl.Invocation;
import cvl.Iterator;
import cvl.IteratorResolution;
import cvl.ResolutionElement;
import cvl.Type;
import cvl.VariabilitySpecification;

public class CVLResolutionModelManager {
	private List<String> errorList = new ArrayList<String>();
	
	
	public List<String> getErrorList() {
		return errorList;
	}

	//method for API
	public CVLModel runResolutionBuilder(List<EObject> objects){
		return runResolutionBuilder(objects, "", null);
	}
	
	//method for API
	public CVLModel runResolutionBuilder(List<EObject> objects, String name){
		return runResolutionBuilder(objects, name, null);
	}
	
	//method for API
	public CVLModel runResolutionBuilder(List<EObject> objects, List<String> errorMsg){
		return runResolutionBuilder(objects, "", errorMsg);
	}
	
	public CVLModel runResolutionBuilder(List<EObject> objects, String name, List<String> errorMsg){
		return runResolutionBuilder(objects, name, errorMsg, null);
	}
	
	/**
	 * Build a resolution model and append it to the CVL model.
	 * @param objects List of executable primitives that should be included in the resolution model.
	 * @param name Name of the resolution model to be generated.
	 * @param errorMsg List to store error messages.
	 * @return CVL model with the generated resolution model appended.
	 */
	public CVLModel runResolutionBuilder(List<EObject> objects, String name, List<String> errorMsg, EditingDomain eDomain){
		CVLModel cm = null;
		errorList.clear();
		cm = getCVLModel(objects.get(0));
		ResolutionElement top = buildResolution(objects, name); 

		if (top != null && errorList.size() == 0){
			if (eDomain != null){
				AddCommand cmd = new AddCommand(eDomain, cm, CvlPackage.eINSTANCE.getCVLModel_ResolutionSpecification(), top);
				eDomain.getCommandStack().execute(cmd);
			}else{
				cm.getResolutionSpecification().add(top);
			}
		}else{
			errorMsg.addAll(errorList);
		}
		
		return cm;
	}
	
	
	private ResolutionElement buildResolution(List<EObject> objects, String name){
		if (checkPath(objects)){
			addElements(objects);
			CVLModel cm = getCVLModel(objects.get(0));
		
			ResolutionElement resElem = buildResolution(cm.getVariabilitySpecification(), objects);
			
			if (resElem != null){
				resElem.setName(name);
			}
			
			return resElem;
		}
		return null; 
	}
	
	private ResolutionElement buildResolution(VariabilitySpecification varSpec, List<EObject> objects){
		
		Stack<VariabilitySpecification> toProcess = new Stack<VariabilitySpecification>();

		ResolutionElement top = CvlFactory.eINSTANCE.createResolutionElement();

		toProcess.push(varSpec);

		while (!toProcess.isEmpty()){
			VariabilitySpecification vs = toProcess.pop();

			if (vs instanceof Iterator){
				top.getResolution().add(buildResolution((Iterator) vs, objects));
			}else if (vs instanceof CompositeVariability){
				CompositeVariability cv = (CompositeVariability) vs;
				toProcess.addAll(0,cv.getVariabilitySpecification()); //depth-first
			}else if (vs instanceof Invocation){
				Invocation inv = (Invocation) vs;
				Type t = inv.getType();
				toProcess.add(0,t.getBody());
			}else{
				//do nothing
			}
		}

		return top;
		
	}

	private IteratorResolution buildResolution(Iterator iterator, List<EObject> objects){

		//create iterator resolution
		IteratorResolution ir = CvlFactory.eINSTANCE.createIteratorResolution();
		ir.setIterator(iterator);

		if (objects.contains(iterator)){

			for (VariabilitySpecification vs : iterator.getVariabilitySpecification()) {
				if (vs instanceof ExecutablePrimitive){
					ExecutablePrimitive prim = (ExecutablePrimitive) vs;
					ResolutionElement re = null;

					if (objects.contains(prim)){
						re = CvlFactory.eINSTANCE.createResolutionElement();
						re.setElement(prim);
						ir.getChoice().add(re);

						if (prim instanceof Iterator){  
							Iterator it = (Iterator) prim;
							IteratorResolution itr = buildResolution(it, objects);
							re.getResolution().add(itr);

						}else if (prim instanceof CompositeVariability){
							CompositeVariability cv = (CompositeVariability) prim;

							for (VariabilitySpecification element : cv.getVariabilitySpecification()) {
								if (element instanceof Iterator){
									Iterator it = (Iterator) element;
									IteratorResolution itr = buildResolution(it, objects);
									re.getResolution().add(itr);
								}else if (element instanceof Invocation){
									Invocation inv = (Invocation) element;
									Type t = inv.getType();
									CompositeVariability c = t.getBody();

									for (VariabilitySpecification elem : c.getVariabilitySpecification()) {
										if (elem instanceof Iterator){
											Iterator it = (Iterator) elem;
											IteratorResolution itr = buildResolution(it, objects);
											re.getResolution().add(itr);
										}
									}
								}else if (element instanceof CompositeVariability){
									ResolutionElement reTemp = buildResolution(element, objects);
									
									if (reTemp != null){
										if (reTemp.getResolution().size() > 0){
											re.getResolution().addAll(reTemp.getResolution());
										}
									}
									
								}
							}

						}else if (prim instanceof Invocation){
							Invocation inv = (Invocation) prim;
							Type t = inv.getType();
							CompositeVariability cv = t.getBody();

							for (VariabilitySpecification element : cv.getVariabilitySpecification()) {
								if (element instanceof Iterator){
									Iterator it = (Iterator) element;
									IteratorResolution itr = buildResolution(it, objects);
									re.getResolution().add(itr);
								}else if (element instanceof CompositeVariability){
									ResolutionElement reTemp = buildResolution(element, objects);
									
									if (reTemp != null){
										if (reTemp.getResolution().size() > 0){
											re.getResolution().addAll(reTemp.getResolution());
										}
									}
									
								}
							}

						}

					}
				}
			}
		}else {
			if (iterator.getLowerLimit() > 0){ //with proper handling of checkPath, this step may be skipped
				errorList.add("Iterator " + iterator.getName() + " need a minimum of " + iterator.getLowerLimit() + " model elements selected.");
			}
		}

		return ir;
	}




	private boolean checkPath(List<EObject> objects){
		//invoke alloy?
		return true;
	}


	private void addElements(List<EObject> objects){
		List<EObject> toProcess = new ArrayList<EObject>();
		toProcess.addAll(objects);

		for (EObject eObject : toProcess) {
			EObject e = eObject;

			while (e.eContainer() != null){
				e = e.eContainer();
				if (e instanceof Type){

				}

				if (!objects.contains(e)){
					objects.add(e);
				}
			}
		}
	}

	private CVLModel getCVLModel(EObject obj){
		while (obj.eContainer() != null){
			obj = obj.eContainer();
		}
		if (obj instanceof CVLModel){
			return (CVLModel) obj;
		}
		return null;
	}
}
